@@ -39,21 +39,33 @@ impl Term {
 
 #[derive(Debug)]
 pub struct ParseOptions {
+    allow_backslash_quotes: bool,
     allow_unicode_escapes: bool,
 }
 
 impl ParseOptions {
     pub fn default() -> Self {
         Self {
+            allow_backslash_quotes: true,
             allow_unicode_escapes: true,
         }
     }
 
+    /// Allows single- and double-quote characters to be escaped inside quoted strings by
+    /// preceding them with a backslash. Similarly, backslash characters themselves can be
+    /// escaped inside quoted strings by preceding them with a backslash. Note that if a
+    /// backslash is followed by any another character, or is inside a non-quoted term, then
+    /// it is treated as a literal backslash.
+    pub fn allow_backslash_quotes(&mut self, allow: bool) -> &mut ParseOptions {
+        self.allow_backslash_quotes = allow;
+        self
+    }
+
     /// Allows `\uXXXXXX` unicode escapes in term values. Each X must be a hexadecimal
     /// character, and all six are required (use zero-padding as needed). To represent
     /// a string like `\u001234` literally in a query string, one or more of the
     /// characters can be escaped via the unicode escape sequence; in this example
-    /// `\u005Cu001234` can be used to escape the initial backslash character
+    /// `\u00005Cu001234` can be used to escape the initial backslash character
     /// with the equivalent unicode escape.
     pub fn allow_unicode_escapes(&mut self, allow: bool) -> &mut ParseOptions {
         self.allow_unicode_escapes = allow;
@@ -78,18 +90,26 @@ enum ParseState {
     Negated, // term started with a '-' character
     SingleQuote, // term started with a single quote
     DoubleQuote, // term started with a double quote
+    SingleQuoteEscape, // encountered backslash inside a single-quoted term
+    DoubleQuoteEscape, // encountered backslash inside a double-quoted term
     RawToken, // term started without quoting
     NegatedSingleQuote, // term started with a '-' followed by a single quote
     NegatedDoubleQuote, // term started with a '-' followed by a double quote
+    NegatedSingleQuoteEscape, // encountered backslash inside NegatedSingleQuote
+    NegatedDoubleQuoteEscape, // encountered backslash inside NegatedDoubleQuote
     NegatedRawToken, // term started with a '-' followed by unquoted characters
     Value, // after encountering the ':' to separate key from value
     NegatedValue, // after encountering the ':' to separated negated key from value
     RawValue, // Once the value has been determined to be unquoted
     SingleQuotedValue, // Once the value has been determined to be single-quoted
     DoubleQuotedValue, // Once the value has been determined to be double-quoted
+    SingleQuotedValueEscape, // Encountered backslash inside SingleQuotedValue
+    DoubleQuotedValueEscape, // Encountered backslash inside DoubleQuotedValue
     NegatedRawValue, // Once the value for a negated term has been determined to be unquoted
     NegatedSingleQuotedValue, // Once the value for a negated term has been determined to be single-quoted
     NegatedDoubleQuotedValue, // Once the value for a negated term has been determined to be double-quoted
+    NegatedSingleQuotedValueEscape, // Encountered backslash inside NegatedSingleQuotedValue
+    NegatedDoubleQuotedValueEscape, // Encountered backslash inside NegatedDoubleQuotedValue
 }
 
 impl ParseState {
@@ -98,11 +118,71 @@ impl ParseState {
             Self::Negated |
             Self::NegatedSingleQuote |
             Self::NegatedDoubleQuote |
+            Self::NegatedSingleQuoteEscape |
+            Self::NegatedDoubleQuoteEscape |
             Self::NegatedRawToken |
             Self::NegatedValue |
             Self::NegatedRawValue |
             Self::NegatedSingleQuotedValue |
-            Self::NegatedDoubleQuotedValue => true,
+            Self::NegatedDoubleQuotedValue |
+            Self::NegatedSingleQuotedValueEscape |
+            Self::NegatedDoubleQuotedValueEscape => true,
+            _ => false,
+        }
+    }
+
+    fn is_single_quote(&self) -> bool {
+        match self {
+            Self::SingleQuote |
+            Self::SingleQuoteEscape |
+            Self::NegatedSingleQuote |
+            Self::NegatedSingleQuoteEscape |
+            Self::SingleQuotedValue |
+            Self::SingleQuotedValueEscape |
+            Self::NegatedSingleQuotedValue |
+            Self::NegatedSingleQuotedValueEscape => true,
+            _ => false,
+        }
+    }
+
+    fn escape(&self) -> Self {
+        match self {
+            ParseState::SingleQuote => ParseState::SingleQuoteEscape,
+            ParseState::DoubleQuote => ParseState::DoubleQuoteEscape,
+            ParseState::NegatedSingleQuote => ParseState::NegatedSingleQuoteEscape,
+            ParseState::NegatedDoubleQuote => ParseState::NegatedDoubleQuoteEscape,
+            ParseState::SingleQuotedValue => ParseState::SingleQuotedValueEscape,
+            ParseState::DoubleQuotedValue => ParseState::DoubleQuotedValueEscape,
+            ParseState::NegatedSingleQuotedValue => ParseState::NegatedSingleQuotedValueEscape,
+            ParseState::NegatedDoubleQuotedValue => ParseState::NegatedDoubleQuotedValueEscape,
+            _ => panic!("Unescapable state"),
+        }
+    }
+
+    fn unescape(&self) -> Self {
+        match self {
+            ParseState::SingleQuoteEscape => ParseState::SingleQuote,
+            ParseState::DoubleQuoteEscape => ParseState::DoubleQuote,
+            ParseState::NegatedSingleQuoteEscape => ParseState::NegatedSingleQuote,
+            ParseState::NegatedDoubleQuoteEscape => ParseState::NegatedDoubleQuote,
+            ParseState::SingleQuotedValueEscape => ParseState::SingleQuotedValue,
+            ParseState::DoubleQuotedValueEscape => ParseState::DoubleQuotedValue,
+            ParseState::NegatedSingleQuotedValueEscape => ParseState::NegatedSingleQuotedValue,
+            ParseState::NegatedDoubleQuotedValueEscape => ParseState::NegatedDoubleQuotedValue,
+            _ => panic!("Unescaped state"),
+        }
+    }
+
+    fn is_escaped(&self) -> bool {
+        match self {
+            Self::SingleQuoteEscape |
+            Self::DoubleQuoteEscape |
+            Self::NegatedSingleQuoteEscape |
+            Self::NegatedDoubleQuoteEscape |
+            Self::SingleQuotedValueEscape |
+            Self::DoubleQuotedValueEscape |
+            Self::NegatedSingleQuotedValueEscape |
+            Self::NegatedDoubleQuotedValueEscape => true,
             _ => false,
         }
     }
@@ -239,14 +319,30 @@ fn parse_terms(raw: &str, opts: &ParseOptions) -> Vec<Term> {
 
             // [Negated] Single/Double quoted state handlers
             (ParseState::SingleQuote, None) |
-            (ParseState::NegatedSingleQuote, None) => {
-                result.push(Term::from_value(format!("{}'{}", if state.is_negated() { "-" } else { "" }, opts.decode_unicode(token))));
+            (ParseState::DoubleQuote, None) |
+            (ParseState::SingleQuoteEscape, None) |
+            (ParseState::DoubleQuoteEscape, None) |
+            (ParseState::NegatedSingleQuote, None) |
+            (ParseState::NegatedDoubleQuote, None) |
+            (ParseState::NegatedSingleQuoteEscape, None) |
+            (ParseState::NegatedDoubleQuoteEscape, None) => {
+                result.push(Term::new(state.is_negated(), None, format!(
+                    "{}{}{}",
+                    if state.is_single_quote() { "'" } else { "\"" },
+                    opts.decode_unicode(token),
+                    if state.is_escaped() { "\\" } else { "" },
+                )));
                 break;
             }
-            (ParseState::DoubleQuote, None) |
-            (ParseState::NegatedDoubleQuote, None) => {
-                result.push(Term::from_value(format!("{}\"{}", if state.is_negated() { "-" } else { "" }, opts.decode_unicode(token))));
-                break;
+            (ParseState::SingleQuoteEscape, Some(ref ch)) |
+            (ParseState::DoubleQuoteEscape, Some(ref ch)) |
+            (ParseState::NegatedSingleQuoteEscape, Some(ref ch)) |
+            (ParseState::NegatedDoubleQuoteEscape, Some(ref ch)) => {
+                if !(*ch == '\'' || *ch == '"' || *ch == '\\') {
+                    token.push('\\');
+                }
+                token.push(*ch);
+                state = state.unescape();
             }
             (ParseState::SingleQuote, Some('\'')) |
             (ParseState::DoubleQuote, Some('"')) |
@@ -260,7 +356,11 @@ fn parse_terms(raw: &str, opts: &ParseOptions) -> Vec<Term> {
             (ParseState::DoubleQuote, Some(ref ch)) |
             (ParseState::NegatedSingleQuote, Some(ref ch)) |
             (ParseState::NegatedDoubleQuote, Some(ref ch)) => {
-                token.push(*ch);
+                if opts.allow_backslash_quotes && *ch == '\\' {
+                    state = state.escape();
+                } else {
+                    token.push(*ch);
+                }
             }
 
             // Raw token state handlers
@@ -341,14 +441,30 @@ fn parse_terms(raw: &str, opts: &ParseOptions) -> Vec<Term> {
             }
 
             (ParseState::SingleQuotedValue, None) |
-            (ParseState::NegatedSingleQuotedValue, None) => {
-                result.push(Term::new(state.is_negated(), key, format!("'{}", opts.decode_unicode(token))));
+            (ParseState::DoubleQuotedValue, None) |
+            (ParseState::SingleQuotedValueEscape, None) |
+            (ParseState::DoubleQuotedValueEscape, None) |
+            (ParseState::NegatedSingleQuotedValue, None) |
+            (ParseState::NegatedDoubleQuotedValue, None) |
+            (ParseState::NegatedSingleQuotedValueEscape, None) |
+            (ParseState::NegatedDoubleQuotedValueEscape, None) => {
+                result.push(Term::new(state.is_negated(), key, format!(
+                    "{}{}{}",
+                    if state.is_single_quote() { "'" } else { "\"" },
+                    opts.decode_unicode(token),
+                    if state.is_escaped() { "\\" } else { "" },
+                )));
                 break;
             }
-            (ParseState::DoubleQuotedValue, None) |
-            (ParseState::NegatedDoubleQuotedValue, None) => {
-                result.push(Term::new(state.is_negated(), key, format!("\"{}", opts.decode_unicode(token))));
-                break;
+            (ParseState::SingleQuotedValueEscape, Some(ref ch)) |
+            (ParseState::DoubleQuotedValueEscape, Some(ref ch)) |
+            (ParseState::NegatedSingleQuotedValueEscape, Some(ref ch)) |
+            (ParseState::NegatedDoubleQuotedValueEscape, Some(ref ch)) => {
+                if !(*ch == '\'' || *ch == '"' || *ch == '\\') {
+                    token.push('\\');
+                }
+                token.push(*ch);
+                state = state.unescape();
             }
             (ParseState::SingleQuotedValue, Some('\'')) |
             (ParseState::DoubleQuotedValue, Some('"')) |
@@ -363,7 +479,11 @@ fn parse_terms(raw: &str, opts: &ParseOptions) -> Vec<Term> {
             (ParseState::DoubleQuotedValue, Some(ref ch)) |
             (ParseState::NegatedSingleQuotedValue, Some(ref ch)) |
             (ParseState::NegatedDoubleQuotedValue, Some(ref ch)) => {
-                token.push(*ch);
+                if opts.allow_backslash_quotes && *ch == '\\' {
+                    state = state.escape();
+                } else {
+                    token.push(*ch);
+                }
             }
         }
     }
@@ -444,9 +564,13 @@ mod tests {
     fn end_unexpectedly() {
         assert_eq!(parse(" -").terms, &[Term::new(false, None, "-")]);
         assert_eq!(parse("'hello").terms, &[Term::new(false, None, "'hello")]);
+        assert_eq!(parse("'hello\\").terms, &[Term::new(false, None, "'hello\\")]);
         assert_eq!(parse("\"hello ").terms, &[Term::new(false, None, "\"hello ")]);
-        assert_eq!(parse("-'hello").terms, &[Term::new(false, None, "-'hello")]);
-        assert_eq!(parse("-\"hello ").terms, &[Term::new(false, None, "-\"hello ")]);
+        assert_eq!(parse("hello\\").terms, &[Term::new(false, None, "hello\\")]);
+        assert_eq!(parse("-'hello").terms, &[Term::new(true, None, "'hello")]);
+        assert_eq!(parse("-'hello\\").terms, &[Term::new(true, None, "'hello\\")]);
+        assert_eq!(parse("-\"hello ").terms, &[Term::new(true, None, "\"hello ")]);
+        assert_eq!(parse("-hello\\").terms, &[Term::new(true, None, "hello\\")]);
 
         assert_eq!(parse("key:\"value ").terms, &[Term::new(false, Some("key"), "\"value ")]);
         assert_eq!(parse("-key:'value ").terms, &[Term::new(true, Some("key"), "'value ")]);
@@ -467,6 +591,58 @@ mod tests {
         assert_eq!(pu("\\u002021:'\\u002022 \\u002023'").terms, &[Term::new(false, Some("\\u002021"), "\u{2022} \u{2023}")]);
     }
 
+    #[test]
+    fn parse_escapes() {
+        let p = |x| parse_with_options(x, ParseOptions::default().allow_backslash_quotes(false));
+        let pe = |x| parse_with_options(x, ParseOptions::default().allow_backslash_quotes(true));
+
+        assert_eq!(pe(r#"'fred says \'hi\''"#).terms, &[Term::new(false, None, "fred says 'hi'")]);
+        assert_eq!(pe(r#""fred says \'hi\'""#).terms, &[Term::new(false, None, "fred says 'hi'")]);
+        assert_eq!(pe(r#""fred says \"hi\"""#).terms, &[Term::new(false, None, "fred says \"hi\"")]);
+        assert_eq!(pe(r#"'fred says \"hi\"'"#).terms, &[Term::new(false, None, "fred says \"hi\"")]);
+        assert_eq!(pe(r#"'backslashes \\ \n'"#).terms, &[Term::new(false, None, "backslashes \\ \\n")]);
+        assert_eq!(pe(r#""backslashes \\ \n""#).terms, &[Term::new(false, None, "backslashes \\ \\n")]);
+
+        assert_eq!(p(r#"'fred says \'hi\''"#).terms, &[Term::new(false, None, "fred says \\"), Term::new(false, None, "hi\\''")]);
+        assert_eq!(p(r#""fred says \'hi\'""#).terms, &[Term::new(false, None, "fred says \\'hi\\'")]);
+        assert_eq!(p(r#""fred says \"hi\"""#).terms, &[Term::new(false, None, "fred says \\"), Term::new(false, None, "hi\\\"\"")]);
+        assert_eq!(p(r#"'fred says \"hi\"'"#).terms, &[Term::new(false, None, "fred says \\\"hi\\\"")]);
+        assert_eq!(p(r#"'backslashes \\ \n'"#).terms, &[Term::new(false, None, "backslashes \\\\ \\n")]);
+        assert_eq!(p(r#""backslashes \\ \n""#).terms, &[Term::new(false, None, "backslashes \\\\ \\n")]);
+
+        assert_eq!(pe(r#"in\"a\\raw\nword"#).terms, &[Term::new(false, None, "in\\\"a\\\\raw\\nword")]);
+        assert_eq!(p(r#"in\"a\\raw\nword"#).terms, &[Term::new(false, None, "in\\\"a\\\\raw\\nword")]);
+
+        assert_eq!(pe(r#"fred:'\'hi\''"#).terms, &[Term::new(false, Some("fred"), "'hi'")]);
+        assert_eq!(pe(r#"fred:"\'hi\'""#).terms, &[Term::new(false, Some("fred"), "'hi'")]);
+        assert_eq!(pe(r#"fred:"\"hi\"""#).terms, &[Term::new(false, Some("fred"), "\"hi\"")]);
+        assert_eq!(pe(r#"fred:'\"hi\"'"#).terms, &[Term::new(false, Some("fred"), "\"hi\"")]);
+        assert_eq!(pe(r#"back:'slashes \\ \n'"#).terms, &[Term::new(false, Some("back"), "slashes \\ \\n")]);
+        assert_eq!(pe(r#"back:"slashes \\ \n""#).terms, &[Term::new(false, Some("back"), "slashes \\ \\n")]);
+
+        assert_eq!(p(r#"fred:'\'hi\''"#).terms, &[Term::new(false, Some("fred"), "\\"), Term::new(false, None, "hi\\''")]);
+        assert_eq!(p(r#"fred:"\'hi\'""#).terms, &[Term::new(false, Some("fred"), "\\'hi\\'")]);
+        assert_eq!(p(r#"fred:"\"hi\"""#).terms, &[Term::new(false, Some("fred"), "\\"), Term::new(false, None, "hi\\\"\"")]);
+        assert_eq!(p(r#"fred:'\"hi\"'"#).terms, &[Term::new(false, Some("fred"), "\\\"hi\\\"")]);
+        assert_eq!(p(r#"back:'slashes \\ \n'"#).terms, &[Term::new(false, Some("back"), "slashes \\\\ \\n")]);
+        assert_eq!(p(r#"back:"slashes \\ \n""#).terms, &[Term::new(false, Some("back"), "slashes \\\\ \\n")]);
+
+        assert_eq!(pe(r#"raw:in\"a\\raw\nword"#).terms, &[Term::new(false, Some("raw"), "in\\\"a\\\\raw\\nword")]);
+        assert_eq!(p(r#"raw:in\"a\\raw\nword"#).terms, &[Term::new(false, Some("raw"), "in\\\"a\\\\raw\\nword")]);
+    }
+
+    #[test]
+    fn parse_unicode_and_escape() {
+        let p = |x| parse_with_options(
+            x,
+            ParseOptions::default().allow_unicode_escapes(true).allow_backslash_quotes(true)
+        );
+
+        assert_eq!(p(r#"fred:"\'sup \u01F436""#).terms, &[Term::new(false, Some("fred"), "'sup \u{1F436}")]);
+        assert_eq!(p(r#"fred:"\'sup \\u01F436""#).terms, &[Term::new(false, Some("fred"), "'sup \u{1F436}")]);
+        assert_eq!(p(r#"fred:"\'sup \u00005cu01F436""#).terms, &[Term::new(false, Some("fred"), "'sup \\u01F436")]);
+    }
+
     #[test]
     fn readme_example() {
         println!("{:?}", parse("from:foo -subject:'a long subject \\u00270c' baz"));
